@@ -54,6 +54,7 @@
 
 #define	CIFS_DEFAULT_KRB5_DIR		"/tmp"
 #define	CIFS_DEFAULT_KRB5_PREFIX	"krb5cc_"
+#define CIFS_DEFAULT_KRB5_KEYTAB	"/etc/krb5.keytab"
 
 #define	MAX_CCNAME_LEN			PATH_MAX + 5
 
@@ -185,6 +186,78 @@ static int krb5cc_filter(const struct dirent *dirent)
 		return 0;
 }
 
+static char *
+init_cc_from_keytab(const char *keytab_name, const char *user)
+{
+	krb5_context context = NULL;
+	krb5_error_code ret;
+	krb5_creds my_creds;
+	krb5_keytab keytab = NULL;
+	krb5_principal me = NULL;
+	krb5_ccache cc = NULL;
+	char *ccname = NULL;
+
+	memset((char *) &my_creds, 0, sizeof(my_creds));
+
+	ret = krb5_init_context(&context);
+	if (ret) {
+		syslog(LOG_DEBUG, "krb5_init_context: %d", (int)ret);
+		goto icfk_cleanup;
+	}
+
+	ret = krb5_kt_resolve(context, keytab_name, &keytab);
+	if (ret) {
+		syslog(LOG_DEBUG, "krb5_kt_resolve: %d", (int)ret);
+		goto icfk_cleanup;
+	}
+
+	ret = krb5_parse_name(context, user, &me);
+	if (ret) {
+		syslog(LOG_DEBUG, "krb5_parse_name: %d", (int)ret);
+		goto icfk_cleanup;
+	}
+
+	ret = krb5_get_init_creds_keytab(context, &my_creds, me,
+			keytab, 0, NULL, NULL);
+	if (ret) {
+		syslog(LOG_DEBUG, "krb5_get_init_creds_keytab: %d", (int)ret);
+		goto icfk_cleanup;
+	}
+
+	ret = krb5_cc_default(context, &cc);
+	if (ret) {
+		syslog(LOG_DEBUG, "krb5_cc_default: %d", (int)ret);
+		goto icfk_cleanup;
+	}
+
+	ret = krb5_cc_initialize(context, cc, me);
+	if (ret) {
+		syslog(LOG_DEBUG, "krb5_cc_initialize: %d", (int)ret);
+		goto icfk_cleanup;
+	}
+
+	ret = krb5_cc_store_cred(context, cc, &my_creds);
+	if (ret)
+		syslog(LOG_DEBUG, "krb5_cc_store_cred: %d", (int)ret);
+
+	ccname = strdup(krb5_cc_default_name(context));
+	if (ccname == NULL)
+		syslog(LOG_ERR, "Unable to allocate memory");
+icfk_cleanup:
+	my_creds.client = 0;
+	krb5_free_cred_contents(context, &my_creds);
+
+	if (me)
+		krb5_free_principal(context, me);
+	if (cc)
+		krb5_cc_close(context, cc);
+	if (keytab)
+		krb5_kt_close(context, keytab);
+	if (context)
+		krb5_free_context(context);
+	return ccname;
+}
+
 /* search for a credcache that looks like a likely candidate */
 static char *find_krb5_cc(const char *dirname, uid_t uid)
 {
@@ -258,9 +331,12 @@ cifs_krb5_get_req(const char *principal, const char *ccname,
 	krb5_keyblock *tokb;
 	krb5_context context;
 	krb5_ccache ccache;
-	krb5_creds in_creds = { }, *out_creds;
+	krb5_creds in_creds, *out_creds;
 	krb5_data apreq_pkt, in_data;
 	krb5_auth_context auth_context = NULL;
+#if defined(HAVE_KRB5_AUTH_CON_SETADDRS) && defined(HAVE_KRB5_AUTH_CON_SET_REQ_CKSUMTYPE)
+	static const uint8_t gss_cksum[24] = { 0x10, 0x00, /* ... */};
+#endif
 
 	ret = krb5_init_context(&context);
 	if (ret) {
@@ -275,6 +351,8 @@ cifs_krb5_get_req(const char *principal, const char *ccname,
 		goto out_free_context;
 	}
 
+	memset(&in_creds, 0, sizeof(in_creds));
+
 	ret = krb5_cc_get_principal(context, ccache, &in_creds.client);
 	if (ret) {
 		syslog(LOG_DEBUG, "%s: unable to get client principal name",
@@ -297,27 +375,76 @@ cifs_krb5_get_req(const char *principal, const char *ccname,
 		goto out_free_principal;
 	}
 
-	apreq_pkt.data = NULL;
 	in_data.length = 0;
+	in_data.data = NULL;
+
+	ret = krb5_auth_con_init(context, &auth_context);
+	if (ret) {
+		syslog(LOG_DEBUG, "%s: unable to create auth_context: %d",
+		       __func__, ret);
+		goto out_free_creds;
+	}
+
+#if defined(HAVE_KRB5_AUTH_CON_SETADDRS) && defined(HAVE_KRB5_AUTH_CON_SET_REQ_CKSUMTYPE)
+	/* Ensure we will get an addressless ticket. */
+	ret = krb5_auth_con_setaddrs(context, auth_context, NULL, NULL);
+	if (ret) {
+		syslog(LOG_DEBUG, "%s: unable to set NULL addrs: %d",
+		       __func__, ret);
+		goto out_free_auth;
+	}
+
+	/*
+	 * Create a GSSAPI checksum (0x8003), see RFC 4121.
+	 *
+	 * The current layout is
+	 *
+	 * 0x10, 0x00, 0x00, 0x00 - length = 16
+	 * 0x00, 0x00, 0x00, 0x00 - channel binding info - 16 zero bytes
+	 * 0x00, 0x00, 0x00, 0x00
+	 * 0x00, 0x00, 0x00, 0x00
+	 * 0x00, 0x00, 0x00, 0x00
+	 * 0x00, 0x00, 0x00, 0x00 - flags
+	 *
+	 * GSS_C_NO_CHANNEL_BINDINGS means 16 zero bytes,
+	 * this is needed to work against some closed source
+	 * SMB servers.
+	 *
+	 * See https://bugzilla.samba.org/show_bug.cgi?id=7890
+	 */
+	in_data.data = discard_const_p(char, gss_cksum);
+	in_data.length = 24;
+	ret = krb5_auth_con_set_req_cksumtype(context, auth_context, 0x8003);
+	if (ret) {
+		syslog(LOG_DEBUG, "%s: unable to set 0x8003 checksum",
+		       __func__);
+		goto out_free_auth;
+	}
+#endif
+
+	apreq_pkt.length = 0;
+	apreq_pkt.data = NULL;
 	ret = krb5_mk_req_extended(context, &auth_context, AP_OPTS_USE_SUBKEY,
 				   &in_data, out_creds, &apreq_pkt);
 	if (ret) {
 		syslog(LOG_DEBUG, "%s: unable to make AP-REQ for %s",
 		       __func__, principal);
-		goto out_free_creds;
+		goto out_free_auth;
 	}
 
 	ret = krb5_auth_con_getsendsubkey(context, auth_context, &tokb);
 	if (ret) {
 		syslog(LOG_DEBUG, "%s: unable to get session key for %s",
 		       __func__, principal);
-		goto out_free_creds;
+		goto out_free_auth;
 	}
 
 	*mechtoken = data_blob(apreq_pkt.data, apreq_pkt.length);
 	*sess_key = data_blob(KRB5_KEY_DATA(tokb), KRB5_KEY_LENGTH(tokb));
 
 	krb5_free_keyblock(context, tokb);
+out_free_auth:
+	krb5_auth_con_free(context, auth_context);
 out_free_creds:
 	krb5_free_creds(context, out_creds);
 out_free_principal:
@@ -389,13 +516,17 @@ handle_krb5_mech(const char *oid, const char *principal, DATA_BLOB * secblob,
 #define DKD_HAVE_IP		0x8
 #define DKD_HAVE_UID		0x10
 #define DKD_HAVE_PID		0x20
+#define DKD_HAVE_CREDUID	0x40
+#define DKD_HAVE_USERNAME	0x80
 #define DKD_MUSTHAVE_SET (DKD_HAVE_HOSTNAME|DKD_HAVE_VERSION|DKD_HAVE_SEC)
 
 struct decoded_args {
 	int ver;
 	char *hostname;
 	char *ip;
+	char *username;
 	uid_t uid;
+	uid_t creduid;
 	pid_t pid;
 	sectype_t sec;
 };
@@ -417,22 +548,45 @@ decode_key_description(const char *desc, struct decoded_args *arg)
 			else
 				len = pos - tkn;
 
-			len -= 4;
+			len -= 5;
 			SAFE_FREE(arg->hostname);
-			arg->hostname = calloc(sizeof(char), len);
-			strlcpy(arg->hostname, tkn + 5, len);
+			arg->hostname = strndup(tkn + 5, len);
+			if (arg->hostname == NULL) {
+				syslog(LOG_ERR, "Unable to allocate memory");
+				return 1;
+			}
 			retval |= DKD_HAVE_HOSTNAME;
+			syslog(LOG_DEBUG, "host=%s", arg->hostname);
 		} else if (!strncmp(tkn, "ip4=", 4) || !strncmp(tkn, "ip6=", 4)) {
 			if (pos == NULL)
 				len = strlen(tkn);
 			else
 				len = pos - tkn;
 
-			len -= 3;
+			len -= 4;
 			SAFE_FREE(arg->ip);
-			arg->ip = calloc(sizeof(char), len);
-			strlcpy(arg->ip, tkn + 4, len);
+			arg->ip = strndup(tkn + 4, len);
+			if (arg->ip == NULL) {
+				syslog(LOG_ERR, "Unable to allocate memory");
+				return 1;
+			}
 			retval |= DKD_HAVE_IP;
+			syslog(LOG_DEBUG, "ip=%s", arg->ip);
+		} else if (strncmp(tkn, "user=", 5) == 0) {
+			if (pos == NULL)
+				len = strlen(tkn);
+			else
+				len = pos - tkn;
+
+			len -= 5;
+			SAFE_FREE(arg->username);
+			arg->username = strndup(tkn + 5, len);
+			if (arg->username == NULL) {
+				syslog(LOG_ERR, "Unable to allocate memory");
+				return 1;
+			}
+			retval |= DKD_HAVE_USERNAME;
+			syslog(LOG_DEBUG, "user=%s", arg->username);
 		} else if (strncmp(tkn, "pid=", 4) == 0) {
 			errno = 0;
 			arg->pid = strtol(tkn + 4, NULL, 0);
@@ -440,9 +594,9 @@ decode_key_description(const char *desc, struct decoded_args *arg)
 				syslog(LOG_ERR, "Invalid pid format: %s",
 				       strerror(errno));
 				return 1;
-			} else {
-				retval |= DKD_HAVE_PID;
 			}
+			syslog(LOG_DEBUG, "pid=%u", arg->pid);
+			retval |= DKD_HAVE_PID;
 		} else if (strncmp(tkn, "sec=", 4) == 0) {
 			if (strncmp(tkn + 4, "krb5", 4) == 0) {
 				retval |= DKD_HAVE_SEC;
@@ -451,6 +605,7 @@ decode_key_description(const char *desc, struct decoded_args *arg)
 				retval |= DKD_HAVE_SEC;
 				arg->sec = MS_KRB5;
 			}
+			syslog(LOG_DEBUG, "sec=%d", arg->sec);
 		} else if (strncmp(tkn, "uid=", 4) == 0) {
 			errno = 0;
 			arg->uid = strtol(tkn + 4, NULL, 16);
@@ -458,9 +613,19 @@ decode_key_description(const char *desc, struct decoded_args *arg)
 				syslog(LOG_ERR, "Invalid uid format: %s",
 				       strerror(errno));
 				return 1;
-			} else {
-				retval |= DKD_HAVE_UID;
 			}
+			retval |= DKD_HAVE_UID;
+			syslog(LOG_DEBUG, "uid=%u", arg->uid);
+		} else if (strncmp(tkn, "creduid=", 8) == 0) {
+			errno = 0;
+			arg->creduid = strtol(tkn + 8, NULL, 16);
+			if (errno != 0) {
+				syslog(LOG_ERR, "Invalid creduid format: %s",
+				       strerror(errno));
+				return 1;
+			}
+			retval |= DKD_HAVE_CREDUID;
+			syslog(LOG_DEBUG, "creduid=%u", arg->creduid);
 		} else if (strncmp(tkn, "ver=", 4) == 0) {	/* if version */
 			errno = 0;
 			arg->ver = strtol(tkn + 4, NULL, 16);
@@ -468,9 +633,9 @@ decode_key_description(const char *desc, struct decoded_args *arg)
 				syslog(LOG_ERR, "Invalid version format: %s",
 				       strerror(errno));
 				return 1;
-			} else {
-				retval |= DKD_HAVE_VERSION;
 			}
+			retval |= DKD_HAVE_VERSION;
+			syslog(LOG_DEBUG, "ver=%d", arg->ver);
 		}
 		if (pos == NULL)
 			break;
@@ -584,12 +749,12 @@ static int ip_to_fqdn(const char *addrstr, char *host, size_t hostlen)
 
 static void usage(void)
 {
-	syslog(LOG_INFO, "Usage: %s [-t] [-v] key_serial", prog);
-	fprintf(stderr, "Usage: %s [-t] [-v] key_serial\n", prog);
+	fprintf(stderr, "Usage: %s [-t] [-v] [-l] key_serial\n", prog);
 }
 
 const struct option long_options[] = {
 	{"trust-dns", 0, NULL, 't'},
+	{"legacy-uid", 0, NULL, 'l'},
 	{"version", 0, NULL, 'v'},
 	{NULL, 0, NULL, 0}
 };
@@ -603,17 +768,20 @@ int main(const int argc, char *const argv[])
 	size_t datalen;
 	unsigned int have;
 	long rc = 1;
-	int c, try_dns = 0;
+	int c, try_dns = 0, legacy_uid = 0;
 	char *buf, *princ = NULL, *ccname = NULL;
 	char hostbuf[NI_MAXHOST], *host;
-	struct decoded_args arg = { };
+	struct decoded_args arg;
 	const char *oid;
+	uid_t uid;
+	char *keytab_name = CIFS_DEFAULT_KRB5_KEYTAB;
 
 	hostbuf[0] = '\0';
+	memset(&arg, 0, sizeof(arg));
 
 	openlog(prog, 0, LOG_DAEMON);
 
-	while ((c = getopt_long(argc, argv, "ctv", long_options, NULL)) != -1) {
+	while ((c = getopt_long(argc, argv, "cltv", long_options, NULL)) != -1) {
 		switch (c) {
 		case 'c':
 			/* legacy option -- skip it */
@@ -621,6 +789,9 @@ int main(const int argc, char *const argv[])
 		case 't':
 			try_dns++;
 			break;
+		case 'l':
+			legacy_uid++;
+			break;
 		case 'v':
 			printf("version: %s\n", VERSION);
 			goto out;
@@ -677,16 +848,28 @@ int main(const int argc, char *const argv[])
 		goto out;
 	}
 
-	if (have & DKD_HAVE_UID) {
-		rc = setuid(arg.uid);
-		if (rc == -1) {
-			syslog(LOG_ERR, "setuid: %s", strerror(errno));
-			goto out;
-		}
-
-		ccname = find_krb5_cc(CIFS_DEFAULT_KRB5_DIR, arg.uid);
+	if (!legacy_uid && (have & DKD_HAVE_CREDUID))
+		uid = arg.creduid;
+	else if (have & DKD_HAVE_UID)
+		uid = arg.uid;
+	else {
+		/* no uid= or creduid= parm -- something is wrong */
+		syslog(LOG_ERR, "No uid= or creduid= parm specified");
+		rc = 1;
+		goto out;
 	}
 
+	rc = setuid(uid);
+	if (rc == -1) {
+		syslog(LOG_ERR, "setuid: %s", strerror(errno));
+		goto out;
+	}
+	ccname = find_krb5_cc(CIFS_DEFAULT_KRB5_DIR, uid);
+
+	/* Couldn't find credcache? Try to use keytab */
+	if (ccname == NULL && arg.username != NULL)
+		ccname = init_cc_from_keytab(keytab_name, arg.username);
+
 	host = arg.hostname;
 
 	// do mech specific authorization
@@ -783,6 +966,7 @@ out:
 	SAFE_FREE(ccname);
 	SAFE_FREE(arg.hostname);
 	SAFE_FREE(arg.ip);
+	SAFE_FREE(arg.username);
 	SAFE_FREE(keydata);
 	return rc;
 }
